# -*- encoding: utf-8 -*-

"""定义 Http Response 要包含的字段信息
"""

from dataclasses import dataclass, asdict
from typing import Any


@dataclass
class ResponseEntity(object):
    """ResponseEntity 用于规范所有 http 响应的数据格式"""

    # 业务层面的消息
    message: str = None

    # 代码层面的异常消息
    error: str = None

    # 返回的数据
    data: Any = None

    def to_dict(self):
        return asdict(self)



# -*- coding: utf-8 -*-

"""dbm-agent 安装

作者: 蒋乐兴|neeky@live.com
时间: 2023-03
"""

import shutil
import logging
import atexit
from pathlib import Path
from dbma.bil.osuser import DBMAUser
from dbma.core.configs import DBM_AGENT_BASE_DIR, DBMAgentConfig
from dbma.bil.net import get_ip_by_card_name
from dbma.core.exception import NetCardNotExistsException


def has_net_card_exists(net_card_name: str):
    """
    检查给定的网卡是否存在
    返回 True 说明给定的网卡存在，如果网卡不存在那么就会报异常 NetCardNotExistsException

    Parameters:
    -----------
    net_card_name: str

    Returns:
    --------
    bool

    Exceptions:
    -----------
    NetCardNotExistsException
    """
    ip = get_ip_by_card_name(net_card_name)
    if ip is None:
        logging.error("not find any ip on {}".format(net_card_name))
        raise NetCardNotExistsException(
            "net card '{}' not exists".format(net_card_name)
        )
    return True


def create_dbma_user():
    """
    创建 dbma 用户
    """
    logging.info("prepare create user dbma .")
    dbma_user = DBMAUser()
    dbma_user.create()
    logging.info("create user dbma done .")
    return dbma_user


def create_dbm_agent_and_database_directorys():
    """
    创建 /usr/local/dbm-agent/*
        /database/mysql/*
        /database/redis/*
    这三个目录和它的子目标
    """
    logging.info("prepare create directions .")
    if not DBM_AGENT_BASE_DIR.exists():
        DBM_AGENT_BASE_DIR.mkdir()

    for subdir in ("etc", "pkgs", "logs", "etc/templates"):
        item = DBM_AGENT_BASE_DIR / subdir
        if not item.exists():
            logging.info("go to create dir {} .".format(item))
            item.mkdir()

    # 创建 MySQL+Redis 会用的的一些公共目录
    DATABASE_DIR = Path("/database")
    if not DATABASE_DIR.exists():
        DATABASE_DIR.mkdir()

    for subdir in ("mysql", "mysql/data", "mysql/binlog", "redis"):
        item = DATABASE_DIR / subdir
        if not item.exists():
            logging.info("go to create dir {} .".format(item))
            item.mkdir()

    logging.info("create directions done .")


def init(net_card_name: str, dbm_center_url_prefix: str):
    """安装 dbm-agent"""
    # 配置日志
    logging.basicConfig(
        level=logging.INFO,
        format="[%(asctime)s %(levelname)s] - [%(threadName)s] - [%(pathname)s %(lineno)d line]  ~  %(message)s",
    )
    logging.info("start install dbm-agent .")

    # 检查给定的网卡是否存在
    has_net_card_exists(net_card_name)

    # 创建用户
    dbma_user = create_dbma_user()

    # 创建目录和子目录
    create_dbm_agent_and_database_directorys()

    # # 复制模板文件
    # logging.info("prepare copy template files .")
    # import dbma

    # basedir = Path(dbma.__file__).parent
    # src = basedir / "static/cnfs/"
    # dest = DBM_AGENT_BASE_DIR / "etc/templates"
    # shutil.copytree(src, dest, dirs_exist_ok=True)
    # logging.info("copy template files done .")

    # 更新配置并保存到磁盘
    dbm_agent_config = DBMAgentConfig()
    dbm_agent_config.host = get_ip_by_card_name(net_card_name)
    dbm_agent_config.dbmcenter_url_prefix = dbm_center_url_prefix
    dbm_agent_config.sync_to_disk()

    ## 权限调整
    dbma_user.chown(DBM_AGENT_BASE_DIR)

    # 给到启动 dbm-agent 的命令提示
    logging.info(
        "install dbm-agent done . \n dbm-agent start \n to start dbm-agent service . "
    )



# -*- coding: utf-8 -*-

"""orchestrator 组件用到的异常对象
"""

from dbma.core.exception import DBMAgentException


class OrchHasBeenInstalledException(DBMAgentException):
    """orch 已经安装过了"""

    pass


class OrchPkgNotExistsException(DBMAgentException):
    """orch 安装包不存在"""

    pass



# -*- coding: utf-8 -*-

"""MySQL 备机相关的操作

作者: 蒋乐兴|neeky@live.com
时间: 2023-03
"""

import time
import logging
from pathlib import Path
from mysql.connector import connect
from dbma.core import messages
from dbma.bil.fun import fname
from dbma.core.configs import dbm_agent_config
from dbma.components.mysql.install import install_mysql
from dbma.components.mysql.commons import get_mysql_version
from dbma.components.mysql.commons import dbma_mysql_cnx


def get_change_source_stmt(version: str = None):
    """根据版本号返回 change master|source 语句

    Parameters:
    -----------
    version: str
        MySQL 版本号

    Return:
    -------
    str
    """
    stmt = "CHANGE REPLICATION SOURCE TO SOURCE_USER='{}', SOURCE_PASSWORD='{}', SOURCE_HOST='{}', SOURCE_PORT={}, SOURCE_AUTO_POSITION=1, SOURCE_SSL=1;"
    if not version.startswith("8.0"):
        stmt = "CHANGE REPLICATION MASTER TO MASTER_USER='{}', MASTER_PASSWORD='{}', MASTER_HOST='{}', MASTER_PORT={}, MASTER_AUTO_POSITION=1, MASTER_SSL=1;"
    return stmt


def get_start_replica_stmt(version: str = None):
    """根据版本号返回 start replica|slave 语句

    Parameters:
    -----------
    version: str
        MySQL 版本号

    Return:
    -------
    str
    """
    stmt = "start replica;"
    if not version.startswith("8.0"):
        stmt = "start slave;"
    return stmt


def change_replica_to(
    port: int = None,
    version: str = None,
    source_ip: str = None,
    source_port: int = None,
):
    """变更复制关系

    Parameters:
    -----------
    port: int
        replica 实例的端口号

    source_host: str
        source 实例的主机号

    source_port: int
        source 实例的端口号

    Returns:
    --------
    None
    """
    logging.info(messages.FUN_STARTS.format(fname()))

    stmt = get_change_source_stmt(version)
    stmt = stmt.format(
        dbm_agent_config.mysql_repl_user,
        dbm_agent_config.mysql_repl_password,
        source_ip,
        source_port,
    )
    with dbma_mysql_cnx(port) as cursor:
        cursor.execute(stmt)
        logging.info(stmt)

    logging.info(messages.FUN_ENDS.format(fname()))


def start_replica(
    port: int = None,
    version: str = None,
    source_ip: str = None,
    source_port: int = None,
):
    """变更复制关系

    Parameters:
    -----------
    port: int
        replica 实例的端口号

    source_host: str
        source 实例的主机号

    source_port: int
        source 实例的端口号

    Returns:
    --------
    None
    """
    logging.info(messages.FUN_STARTS.format(fname()))

    stmt = get_start_replica_stmt(version)
    with dbma_mysql_cnx(port) as cursor:
        cursor.execute(stmt)
        logging.info(stmt)

    logging.info(messages.FUN_ENDS.format(fname()))


def install_mysql_replica(
    port: int = 3306,
    pkg: Path = None,
    innodb_buffer_pool_size: str = "128M",
    source_ip: str = None,
    source_port: int = None,
):
    """安装 MySQL Replica 结点

    Paramters:
    ----------
    port: int
        Replica 实例的端口号

    pkg: Path
        Replica 实例的安装包

    innodb_buffer_pool_size: str
        innodb_buffer_pool_size 的大小

    source_ip: str
        Source 实例的主机 ip

    source_port: int
        Source 实例的端口
    """
    logging.info(messages.FUN_STARTS.format(fname()))

    version = get_mysql_version(pkg.name)

    # 第一步安装 MySQL Replica 结点
    install_mysql(port, pkg, innodb_buffer_pool_size)

    # MySQL 安装之后要 sleep 一下，希望这个时候 MySQL 的监听已经完成
    seconds = 11
    time.sleep(seconds)

    # 第二步执行 change-master-to
    change_replica_to(port, version, source_ip=source_ip, source_port=source_port)

    # 第三步执行 start-replica
    start_replica(port, version, source_ip=source_ip, source_port=source_port)

    logging.info(messages.FUN_ENDS.format(fname()))



# -*- coding: utf-8 -*-

"""MySQL 配置文件相关的逻辑

作者: 蒋乐兴|neeky@live.com
时间: 2023-03
"""

import re
import os
import json
import logging
import random
from enum import Enum
from pathlib import Path
from jinja2 import Template
from datetime import datetime
from dataclasses import dataclass, asdict
from dbma.core import messages
from dbma.bil.fun import fname
from dbma.core.configs import dbm_agent_config
from dbma.components.mysql.exceptions import MySQLTemplateFileNotExistsException


class MySQLTemplateTypes(Enum):
    """定义配置文件模板类型"""

    # MySQL 配置文件
    MYSQL_CONFIG_FILE = 1

    # MySQL init 专用配置文件
    MYSQL_INIT_CONFIG_FILE = 2

    # MySQL systemd 配置文件
    MYSQL_SYSTEMD_FILE = 3


@dataclass
class MySQLConfig(object):
    """MySQL 配置文件的动态生成"""

    # basic
    basedir: str = None
    port: str = None
    innodb_buffer_pool_size: str = None
    now: str = datetime.now().isoformat()

    # region global-config
    version: str = None
    user: str = None
    datadir: str = None
    server_id: int = None
    bind_address: str = "*"
    admin_address: str = "127.0.0.1"
    admin_port: str = None
    mysqlx_port: str = None
    socket: str = None
    mysqlx_socket: str = None
    pid_file: str = None
    character_set_server: str = "utf8mb4"
    collation_server: str = "utf8mb4_0900_ai_ci"
    collation_database: str = "utf8mb4_0900_ai_ci"
    open_files_limit: int = 102400
    max_prepared_stmt_count: int = 4194304
    skip_name_resolve: str = "ON"
    super_read_only: str = "ON"
    read_only: str = "ON"
    log_timestamps: str = "SYSTEM"
    event_scheduler: str = "ON"
    auto_generate_certs: str = "ON"
    activate_all_roles_on_login: str = "OFF"
    end_markers_in_json: str = "OFF"
    tmpdir: str = "/tmp/"
    max_connections: int = 4096
    thread_cache_size: int = 32
    autocommit: str = "ON"
    sort_buffer_size: int = 262144
    join_buffer_size: int = 262144
    eq_range_index_dive_limit: int = 200
    authentication_policy: str = "*,,"
    show_gipk_in_create_table_and_information_schema: str = "ON"
    sql_generate_invisible_primary_key: str = "ON"

    # endregion global-config

    # region for-tables
    big_tables: str = "ON"
    sql_require_primary_key: str = "ON"
    lower_case_table_names: int = 1
    auto_increment_increment: int = 1
    auto_increment_offset: int = 1
    table_open_cache: int = 10240
    table_definition_cache: int = 2048
    table_open_cache_instances: int = 8
    # endregion for-tables

    # region for-nets
    max_allowed_packet: str = "1G"
    connect_timeout: int = 10
    interactive_timeout: int = 28800
    net_read_timeout: int = 30
    net_retry_count: int = 10
    net_write_timeout: int = 60
    net_buffer_length: int = 32768
    # endregion for-nets

    # region general-logs
    log_output: str = "FILE"
    general_log: str = "OFF"
    general_log_file: str = "general.log"
    # endregion general-logs

    # region slow-logs
    slow_query_log: str = "ON"
    slow_query_log_file: str = "slow.log"
    long_query_time: str = 1
    log_queries_not_using_indexes: str = "OFF"
    log_slow_admin_statements: str = "ON"
    log_slow_replica_statements: str = "ON"
    log_slow_slave_statements: str = "ON"
    # endregion slow-logs

    # region error-logs
    log_error: str = "err.log"
    log_statements_unsafe_for_binlog: str = "ON"
    # endregion error-logs

    # region binlogs
    log_bin: str = None
    binlog_checksum: str = "CRC32"
    log_bin_trust_function_creators: str = "ON"
    binlog_direct_non_transactional_updates: str = "OFF"
    expire_logs_days: int = 7
    binlog_expire_logs_seconds: int = 2592000
    binlog_error_action: str = "ABORT_SERVER"
    binlog_format: str = "ROW"
    max_binlog_stmt_cache_size: int = 18446744073709547520
    max_binlog_cache_size: int = 18446744073709547520
    max_binlog_size: str = "1GB"
    binlog_order_commits: str = "ON"
    binlog_row_image: str = "FULL"
    binlog_row_metadata: str = "FULL"
    binlog_rows_query_log_events: str = "ON"
    binlog_stmt_cache_size: int = 32768
    log_replica_updates: str = "ON"
    binlog_transaction_compression: str = "ON"
    binlog_transaction_dependency_history_size: int = 25000
    binlog_transaction_dependency_tracking: str = "WRITESET"
    transaction_write_set_extraction: str = "XXHASH64"
    sync_binlog: int = 1
    binlog_cache_size: int = 32768
    binlog_group_commit_sync_delay: int = 0
    binlog_group_commit_sync_no_delay_count: int = 0
    # endregion binlogs

    # region gtids
    gtid_mode: str = "ON"
    binlog_gtid_simple_recovery: str = "ON"
    enforce_gtid_consistency: str = "ON"
    gtid_executed_compression_period: int = 0
    # endregion gtids

    # region clone
    plugin_load_add: str = "mysql_clone.so"
    clone_autotune_concurrency: str = "ON"
    clone_buffer_size: str = "16M"
    clone_block_ddl: str = "OFF"
    clone_delay_after_data_drop: int = 300
    clone_ddl_timeout: int = 300
    clone_donor_timeout_after_network_failure: int = 5
    clone_enable_compression: str = "ON"
    clone_max_concurrency: int = 16
    clone_max_data_bandwidth: int = 0
    clone_max_network_bandwidth: int = 0
    clone_valid_donor_list: str = "NULL"
    # endregion clone

    # region replication
    rpl_semi_sync_master_enabled: str = "ON"
    rpl_semi_sync_slave_enabled: str = "ON"
    rpl_semi_sync_master_timeout: int = 1000
    rpl_semi_sync_master_wait_point: str = "AFTER_SYNC"
    rpl_semi_sync_master_wait_no_slave: str = "ON"
    rpl_semi_sync_master_wait_for_slave_count: int = 1
    sync_source_info: int = 10000
    skip_slave_start: str = "ON"
    skip_replica_start: str = "ON"
    log_slave_updates: str = "ON"
    replica_load_tmpdir: str = "/tmp/"
    master_info_repository: str = "table"
    relay_log_info_repository: str = "table"
    plugin_load_add: str = "semisync_master.so"
    plugin_load_add: str = "semisync_slave.so"
    relay_log: str = "relay-bin"
    sync_relay_log: int = 10000
    sync_relay_log_info: int = 10000
    slave_preserve_commit_order: str = "ON"
    replica_preserve_commit_order: str = "ON"
    slave_parallel_type: str = "LOGICAL_CLOCK"
    replica_parallel_type: str = "LOGICAL_CLOCK"
    slave_parallel_workers: int = 4
    replica_parallel_workers: int = 4
    replica_max_allowed_packet: str = "1G"
    transaction_write_set_extraction: str = "XXHASH64"
    # endregion replication

    # region engines
    default_storage_engine: str = "InnoDB"
    default_tmp_storage_engine: str = "InnoDB"
    internal_tmp_mem_storage_engine: str = "TempTable"
    # endregion engines

    # region innodbs
    innodb_data_home_dir: str = "./"
    innodb_data_file_path: str = "ibdata1:64M:autoextend"
    innodb_page_size: int = 16384
    innodb_default_row_format: str = "DYNAMIC"
    innodb_log_group_home_dir: str = "redo-files"
    innodb_redo_log_capacity: str = "1G"
    innodb_redo_log_encrypt: str = "OFF"
    innodb_online_alter_log_max_size: str = "256M"
    innodb_undo_directory: str = "undo-files"
    innodb_undo_log_encrypt: str = "OFF"
    innodb_undo_log_truncate: str = "ON"
    innodb_max_undo_log_size: str = "1G"
    innodb_rollback_on_timeout: str = "OFF"
    innodb_rollback_segments: int = 128
    innodb_log_checksums: str = "ON"
    innodb_checksum_algorithm: str = "crc32"
    innodb_log_compressed_pages: str = "ON"
    innodb_doublewrite: str = "ON"
    innodb_doublewrite_dir: str = "dblwr-files"
    innodb_doublewrite_files: int = 4
    innodb_doublewrite_batch_size: int = 0
    innodb_commit_concurrency: int = 0
    innodb_read_only: str = "OFF"
    innodb_dedicated_server: str = "OFF"
    innodb_old_blocks_pct: int = 37
    innodb_old_blocks_time: int = 1000
    innodb_random_read_ahead: str = "OFF"
    innodb_read_ahead_threshold: int = 56
    innodb_max_dirty_pages_pct_lwm: int = 10
    innodb_max_dirty_pages_pct: int = 90
    innodb_lru_scan_depth: str = 1024
    innodb_adaptive_flushing: str = "ON"
    innodb_adaptive_flushing_lwm: int = 10
    innodb_flushing_avg_loops: int = 30
    innodb_buffer_pool_dump_pct: int = 25
    innodb_buffer_pool_dump_at_shutdown: str = "ON"
    innodb_buffer_pool_load_at_startup: str = "ON"
    innodb_buffer_pool_filename: str = "ib_buffer_pool"
    innodb_stats_persistent: str = "ON"
    innodb_stats_on_metadata: str = "OFF"
    innodb_stats_method: str = "nulls_equal"
    innodb_stats_auto_recalc: str = "ON"
    innodb_stats_include_delete_marked: str = "OFF"
    innodb_stats_persistent_sample_pages: int = 20
    innodb_stats_transient_sample_pages: int = 8
    innodb_status_output: str = "OFF"
    innodb_status_output_locks: str = "OFF"
    innodb_buffer_pool_dump_now: str = "OFF"
    innodb_buffer_pool_load_abort: str = "OFF"
    innodb_buffer_pool_load_now: str = "OFF"
    innodb_thread_concurrency: int = 0
    innodb_concurrency_tickets: int = 5000
    innodb_thread_sleep_delay: int = 10000
    innodb_adaptive_max_sleep_delay: int = 150000
    innodb_read_io_threads: int = 4
    innodb_write_io_threads: int = 4
    innodb_use_native_aio: str = "ON"
    innodb_flush_sync: str = "ON"
    innodb_spin_wait_delay: int = 6
    innodb_purge_threads: int = 4
    innodb_purge_batch_size: int = 300
    innodb_purge_rseg_truncate_frequency: int = 128
    innodb_deadlock_detect: str = "ON"
    innodb_print_all_deadlocks: str = "OFF"
    innodb_lock_wait_timeout: int = 50
    innodb_table_locks: str = "ON"
    innodb_sync_array_size: int = 1
    innodb_sync_spin_loops: int = 30
    innodb_print_ddl_logs: str = "OFF"
    innodb_replication_delay: int = 0
    innodb_cmp_per_index_enabled: str = "OFF"
    innodb_disable_sort_file_cache: str = "OFF"
    innodb_numa_interleave: str = "OFF"
    innodb_strict_mode: str = "ON"
    innodb_sort_buffer_size: int = 1048576
    innodb_fast_shutdown: int = 0
    innodb_force_load_corrupted: str = "OFF"
    innodb_force_recovery: str = 0
    innodb_temp_tablespaces_dir: str = "innodb-temp-tablespaces"
    innodb_tmpdir: str = "innodb-temps"
    innodb_temp_data_file_path: str = "ibtmp1:12M:autoextend"
    innodb_page_cleaners: int = 4
    innodb_adaptive_hash_index: str = "OFF"
    innodb_adaptive_hash_index_parts: int = 8
    innodb_flush_log_at_timeout: int = 1
    innodb_fsync_threshold: int = 0
    innodb_fill_factor: int = 96
    innodb_file_per_table: str = "ON"
    innodb_autoextend_increment: int = 64
    innodb_open_files: int = 102400
    innodb_buffer_pool_chunk_size: str = "128M"
    innodb_flush_neighbors: int = 0
    innodb_io_capacity: int = 2000
    innodb_io_capacity_max: int = 20000
    innodb_autoinc_lock_mode: int = 2
    innodb_change_buffer_max_size: int = 25
    innodb_flush_method: str = "fsync"
    innodb_change_buffering: str = "ALL"
    innodb_flush_log_at_trx_commit: int = 1
    innodb_buffer_pool_instances: int = None
    innodb_log_buffer_size: str = None

    # expire
    innodb_log_files_in_group: int = 8
    innodb_log_file_size: str = "128M"
    # endregion innodbs

    # region ps
    performance_schema: str = "ON"
    # endregion ps

    def calcu_second_attrs(self):
        """根据已有的配置推导出相关的其它配置"""
        logging.info(messages.FUN_STARTS.format(fname()))

        self._calcu_deps_port()
        self._calcu_random_attrs()
        self._calcu_deps_mem()
        self._calcu_deps_basedir()

        logging.info(messages.FUN_ENDS.format(fname()))

    def save_to_target_dir(self, target_dir: Path = None):
        """以 json 格式保存到目标目录

        Parameters:
        -----------
        target_dir: Path
            json 格式配置文件要保存到的路径(parent-dir)
        """
        logging.info(messages.FUN_STARTS.format(fname()))

        # 检查参数
        if target_dir is None:
            # 没有给出 target_dir 就保存到 datadir
            json_file = Path(self.datadir) / "mysql-config.json"
        else:
            json_file = Path(target_dir) / "mysql-config.json"

        # 检查目录是否存在
        if not Path(target_dir).exists():
            logging.error(
                "dir '{}' not exists, skip save config to it .".format(target_dir)
            )
            logging.info(messages.FUN_ENDS.format(fname()))
            return

        # 保存到给定目录
        logging.info("write config file to '{}' .".format(json_file))
        json_data = asdict(self)
        with open(json_file, "w") as f:
            f.write(json.dumps(json_data, indent=4))

        logging.info(messages.FUN_ENDS.format(fname()))

    def find_mysql_template_file(self, mtt: MySQLTemplateTypes = None):
        """根据版本号加载配置文件模板，如果找不到对应的模板文件就返回 None

        Parameters:
        -----------
        mtt: MySQLTemplateTypes
            MySQL 配置文件模板类型

        Return:
        -------
        Path

        Exceptions:
        -----------
        MySQLTemplateFileNotExistsException

        """
        logging.info(messages.FUN_STARTS.format(fname()))

        # 根据需要的配置文件类型返回模板文件(Path)
        import dbma

        if mtt == MySQLTemplateTypes.MYSQL_CONFIG_FILE:
            template_file = Path(
                dbma.__file__
            ).parent / "static/cnfs/mysql-{}.cnf.jinja".format(self.version)
            # dbm-agent 只对 5.7.x 提供有限的支持、配置文件的模板最高为 5.7.25 、也就是说所有的版本都用这个一个模板
            if self.version.startswith("5.7"):
                template_file = Path(
                    dbma.__file__
                ).parent / "static/cnfs/mysql-{}.cnf.jinja".format("5.7.25")
                logging.info(
                    "5.7.xx version well using config template {}".format(template_file)
                )

        elif mtt == MySQLTemplateTypes.MYSQL_INIT_CONFIG_FILE:
            short_version = "8.0" if self.version.startswith("8.0") else "5.7"
            template_file = Path(
                dbma.__file__
            ).parent / "static/cnfs/mysql-{}-init-only.jinja".format(short_version)
        elif mtt == MySQLTemplateTypes.MYSQL_SYSTEMD_FILE:
            template_file = (
                Path(dbma.__file__).parent / "static/cnfs/mysqld.service.jinja"
            )
        logging.info("using template file {}".format(template_file))

        # 检查一下是否存在
        if template_file.exists():
            logging.info(messages.FUN_ENDS.format(fname()))
            return template_file

        logging.warning("template file '{}' not exists ".format(template_file))
        raise MySQLTemplateFileNotExistsException(template_file)

    def render_mysql_template(self, template: str = None):
        """渲染给定的 template 文件

        Parameters:
        -----------
        template: str
            模板文件的文字内容

        Return:
        -------
        str
            渲染之后的配置文件
        """
        logging.info(messages.FUN_STARTS.format(fname()))
        if template is None:
            logging.error("template is None .")
            return

        if not template.exists():
            logging.error("template not exists .")
            return

        with open(template, "r") as f:
            content = f.read()

        t = Template(content)
        logging.info(messages.FUN_ENDS.format(fname()))
        return t.render(asdict(self))

    def generate_cnf_config_file(self):
        """生成配置文件  /etc/my-{port}-cnf"""
        logging.info(messages.FUN_STARTS.format(fname()))
        # 根据版本号加载配置文件模板
        # 渲染模板
        # 保存渲染后的内容到文件
        try:
            tempate = self.find_mysql_template_file(
                MySQLTemplateTypes.MYSQL_CONFIG_FILE
            )
        except MySQLTemplateFileNotExistsException as err:
            logging.error("cannot generate cnf config file, becuase {}".format(err))
            raise err

        content = self.render_mysql_template(tempate)

        with open(Path("/etc/my-{}.cnf".format(self.port)), "w") as f:
            f.write(content)
        logging.info(messages.FUN_ENDS.format(fname()))

    def generate_init_cnf_config_file(self):
        """生成初始化专用文件, 用于解决 plugin-add 导致的一系列问题

        2023-03-29T11:25:08.320436+08:00 0 [Warning] [MY-013501] [Server] Ignoring --plugin-load[_add] list as the server is running with --initialize(-insecure).
        2023-03-29T11:25:09.326699+08:00 0 [ERROR] [MY-000067] [Server] unknown variable 'rpl_semi_sync_master_enabled=ON'.
        2023-03-29T11:25:09.326717+08:00 0 [ERROR] [MY-013236] [Server] The designated data directory /database/mysql/data/3306/ is unusable. You can remove all files that the server added to it.

        所以我们要在 init 的时候生成一份没有 plugin-add 的版本
        """
        logging.info("starts generate init cnf config file .")
        # 查询 mysql-init 的配置文件模板
        try:
            tempate = self.find_mysql_template_file(
                MySQLTemplateTypes.MYSQL_INIT_CONFIG_FILE
            )
        except MySQLTemplateFileNotExistsException as err:
            logging.error("cannot generate cnf config file, becuase {}".format(err))
            raise err

        content = self.render_mysql_template(tempate)

        with open(Path(dbm_agent_config.mysql_init_cnf), "w") as f:
            f.write(content)

        logging.info("starts generate init cnf config file .")

    def generate_systemd_cnf_config(self):
        """ """
        logging.info("starts generate systemd config file .")
        # 查询 mysql-init 的配置文件模板
        try:
            tempate = self.find_mysql_template_file(
                MySQLTemplateTypes.MYSQL_SYSTEMD_FILE
            )
        except MySQLTemplateFileNotExistsException as err:
            logging.error("cannot generate cnf config file, becuase {}".format(err))
            raise err

        content = self.render_mysql_template(tempate)

        with open(
            Path("/usr/lib/systemd/system/mysqld-{}.service".format(self.port)), "w"
        ) as f:
            f.write(content)

        logging.info("ends generate systemd config file .")

    def _calcu_deps_port(self):
        """计算依赖于 port 的配置项

        user = 'mysql' + str(port)
        """
        logging.info("starts _calcu_dep_port fun . ")
        # 设置参数
        # 目前 json 模块还不支持序列化 Path 对象，所以这里统一用 str 来表示路径
        self.user = dbm_agent_config.mysql_user_prefix + str(self.port)
        self.datadir = os.path.join(
            dbm_agent_config.mysql_datadir_parent, str(self.port)
        )
        self.admin_port = self.port * 10 + 2
        self.mysqlx_port = self.port * 10
        self.log_bin = os.path.join(
            dbm_agent_config.mysql_binlogdir_parent, str(self.port) + "/binlog"
        )

        # 跟据 datadir 计算 socket 文件的位置
        self.socket = os.path.join(self.datadir, "mysql.sock")
        self.mysqlx_socket = os.path.join(self.datadir, "mysqlx.sock")
        self.pid_file = os.path.join(self.datadir, "mysql.pid")

        logging.info("set user to {} .".format(self.user))
        logging.info("set datadir to {} .".format(self.datadir))
        logging.info("set admin_port to {} .".format(self.admin_port))
        logging.info("set mysqlx_port to {} .".format(self.mysqlx_port))
        logging.info("ends _calcu_dep_port fun . ")

    def _calcu_random_attrs(self):
        """生成随机值"""
        self.server_id = random.randint(1024, 8192)

    def _calcu_deps_mem(self):
        """计算内存相关的值"""
        if self.innodb_buffer_pool_size.endswith("M"):
            # M 级别
            self.innodb_buffer_pool_instances = 1
            self.innodb_log_buffer_size = "64M"
        elif self.innodb_buffer_pool_size.endswith("G"):
            # G 级别
            size = re.match(r"\d*", self.innodb_buffer_pool_size).group(0)
            size = int(size)
            if size <= 2:
                self.innodb_buffer_pool_instances = 1
                self.innodb_log_buffer_size = "64M"
            elif size <= 4:
                self.innodb_buffer_pool_instances = size
                self.innodb_log_buffer_size = "128M"
            elif size <= 8:
                self.innodb_buffer_pool_instances = size
                self.innodb_log_buffer_size = "256M"
            elif size <= 16:
                self.innodb_buffer_pool_instances = size
                self.innodb_log_buffer_size = "512M"
            else:
                self.innodb_buffer_pool_instances = 16
                self.innodb_log_buffer_size = "1G"

    def _calcu_deps_basedir(self):
        """根据 basedir 计算出 MySQL 的 version"""
        m = re.search(r"mysql-(?P<version>\d{1}.\d{1,2}.\d{1,2})-linux", self.basedir)
        self.version = m.group("version")



# -*- coding: utf-8 -*-

"""dbm-agent http-server

"""

import os
import logging
from aiohttp import web
from urllib.parse import urlparse
from logging.handlers import RotatingFileHandler
from dbma.bil.daemon import start_daemon, stop_daemon
from dbma.bil.osuser import get_uid_gid, is_root, DBMAUser
from dbma.core.router import routes
from dbma.core.threads import backends
from dbma.core.configs import DBMAgentConfig
from dbma.core.views import dbmagentview as _
from dbma.components.mysql.backends.clears import start_clear_tasks
from dbma.components.mysql.views import defaultsview as _

dbm_agent_config = DBMAgentConfig()
dbm_center_host = urlparse(dbm_agent_config.dbmcenter_url_prefix).hostname


# 是否启动 dbmacenter 安全增强
@web.middleware
async def enforce_token(request: web.Request, handler):
    """验证 dbmcenter_token 值是否正确，如果这个值不对就直接返回 http-511"""
    logging.info("middleware starts enforce_token remoter '{}'".format(request.remote))

    # 只能是 token 对的上我们才处理
    token = request.headers.get("dbmcenter_token", "")
    if token == dbm_agent_config.dbmcenter_token:
        response = await handler(request)
        logging.info("middleware ends enforce_token {}".format(request.remote))
        return response

    message = "enforce_token error , missing dbmcenter_token or dbmcenter_token value error, remote = '{}' ".format(
        request.remote
    )
    # 指示客户端需要进行身份验证才能获得网络访问权限
    logging.warn(message)
    logging.warn("middleware ends enforce_token {}".format(request.remote))
    return web.json_response(
        {"message:": message, "error": message, "data": None}, status=511
    )


def start():
    """ """
    # 检查用户
    if not is_root():
        print("please use root use run this program! ")
        exit(1)

    dbma_user = DBMAUser()
    if not dbma_user.is_exists():
        print("please init dbm-agent before use. ")
        exit(2)

    # 切换到普通用户
    uid, gid = get_uid_gid(dbma_user.name)
    os.setegid(gid)
    os.seteuid(uid)

    # 以服务运行
    # dbm_agent_config = DBMAgentConfig()
    start_daemon(dbm_agent_config.pid_file)
    print("log file '{}' .".format(dbm_agent_config.log_file))

    # 配置日志
    levels = {
        "info": logging.INFO,
        "debug": logging.DEBUG,
        "error": logging.ERROR,
        "warn": logging.WARN,
    }
    handler = RotatingFileHandler(
        filename=dbm_agent_config.log_file,
        maxBytes=128 * 1024 * 1024,
        backupCount=8,
        encoding="utf8",
    )
    logging.basicConfig(
        handlers=[handler],
        level=levels[dbm_agent_config.log_level],
        format="[%(asctime)s %(levelname)s] - [%(threadName)s] - [%(pathname)s %(lineno)d line]  ~  %(message)s",
    )

    # 服务启动的日志头
    logging.info("-" * 21)
    logging.info("| start dbm-agent . |")
    logging.info("-" * 21)
    logging.info("logging-level {}".format(dbm_agent_config.log_level))

    # 启动后台线程
    logging.info("start backends threads .")
    backends.start_cycle_tasks()
    start_clear_tasks()

    # 启动 http 服务
    logging.info("going to start dbm-agent http-server bind on 0.0.0.0:8086 .")

    app = web.Application(middlewares=[enforce_token])
    app.add_routes(routes)
    web.run_app(app, host="0.0.0.0", port=8086, access_log=None)


def stop():
    """关闭 dbm-agent 服务"""
    dbm_agent_config = DBMAgentConfig()
    stop_daemon(dbm_agent_config.pid_file)



# -*- coding: utf-8 -*-

"""在 MySQL 实例层面相关的函数

作者: 蒋乐兴|neeky@live.com
时间: 2023-03
"""

from pathlib import Path
from mysql.connector import connect
from dbma.core.configs import dbm_agent_config


def is_instance_exists(port: int = None):
    """检查给定端口的实例在当前的机器上是否存在(只要 datadir 存在就算是这个实例存在)

    Paramters:
    ----------
    port: int
        MySQL 端口

    Return:
    -------
    bool

    Exceptions:
    -----------
    Exception
    """
    datadir = Path(dbm_agent_config.mysql_datadir_parent) / str(port)
    return datadir.exists()



# -*- coding: utf-8 -*-

"""dbm-agent exceptions

"""


class DBMAgentException(Exception):
    """所有 dbm-agent 异常的基类"""

    pass


# region files


class FileExistsException(DBMAgentException):
    """文件不存在异常"""

    pass


class FileNotExistsException(DBMAgentException):
    """文件已经存在异常"""

    pass


# endregion files


# region directorys
class DirectoryExistsException(DBMAgentException):
    """目录已经存在"""

    pass


class DirectoryNotExistsException(DBMAgentException):
    """目录不存在的异常"""

    pass


# endregion directorys


# region netcards
class NetCardNotExistsException(DBMAgentException):
    """网卡不存在"""

    pass


# endregion netcards



# -*- coding: utf-8 -*-

"""MySQL 组件的公共函数
"""

import re
import time
import logging
import shutil
import contextlib
from enum import Enum
import mysql.connector
from pathlib import Path
from dbma.bil.fun import fname
from dbma.core import messages
from dbma.bil.osuser import MySQLUser
from dbma.bil.cmdexecutor import exe_shell_cmd
from dbma.core.configs import dbm_agent_config
from dbma.components.mysql.exceptions import NotSuportMySQLDirectoryType


default_pkg = Path(
    "/usr/local/dbm-agent/pkgs/mysql-{}-linux-glibc2.28-x86_64.tar.gz".format(
        dbm_agent_config.mysql_default_version
    )
)


class MySQLDirs(Enum):
    DATA = 1
    BINLOG = 2
    BACKUP = 3


def get_mysql_version(pkg_name: str = None):
    """根据 pkg_name 的值，提取 MySQL 的版本号。

    Paramters:
    ----------
    pkg_name: str
        MySQL 安装包的名字 et: mysql-8.0.31-linux-glibc2.12-x86_64.tar.xz

    Return:
    -------
    str | None
    """
    p = re.compile(r"mysql-(?P<version>\d{1,1}.\d{1,1}.\d{1,2})-linux-glibc")
    if (m := p.match(pkg_name)) and m is not None:
        return m.group("version")
    # 异常场景，目前还不清楚最理想的处理方式
    return None


def pkg_to_basedir(pkg: Path = default_pkg):
    """根据 pkg 的名字计算 basedir 的名字

    Parameter:
    ----------
    pkg: Path
        MySQL 安装包的全路径

    Return:
    -------
        Path
    """
    return Path("/usr/local") / (pkg.name.replace(".tar.gz", "").replace(".tar.xz", ""))


@contextlib.contextmanager
def dbma_mysql_cnx(
    port: int = 3306,
    user: str = dbm_agent_config.mysql_dbma_user,
    password: str = dbm_agent_config.mysql_dbma_password,
):
    """建立到本地 MySQL 结点的短连接, 并返回 cursor 对象, 当连接遇到异常的时候返回 None .

    Parameters:
    -----------
    port: int
        MySQL 端口号

    user: str
        连接 MySQL 的用户名

    password: str
        连接 MySQL 的用户密码

    Return:
    -------
    cursor
    """
    cnx = None
    cursor = None
    host = "127.0.0.1"
    logging.info("dbma_mysql_cnx: connect to {}:{} .".format(host, port))
    try:
        # 连接数据库并返回游标
        cnx = mysql.connector.connect(
            host=host, port=port, user=user, password=password, autocommit=True
        )
        cursor = cnx.cursor()
        yield cursor
    except mysql.connector.errors.Error as err:
        # 连接异常的时候返回 None
        logging.error("dbma_mysql_cnx: can't connect to {}:{}".format(host, port))
        yield None
    finally:
        # 资源回收
        try:
            if cursor is not None:
                cursor.close()
            if cnx is not None:
                cnx.close()
        except (mysql.connector.errors.Error, Exception) as err:
            # 如果 cursor 的接收缓冲区还有内容没有读的话，cursor.close() 会报异常 mysql.connector.errors.InternalError: Unread result found
            logging.info("dbma_mysql_cnx: close cnx got error '{}' .".format(err))


def export_cmds_to_path(basedir: Path = None):
    """根据 basedir 设置 PATH 环境变量"""
    # 读出所有的行
    with open("/etc/profile") as f:
        lines = [line for line in f]

    # 检查是否已经导出了
    export_str = "export PATH={}/bin:$PATH\n".format(basedir)
    if export_str in lines:
        logging.info("has exported.")
        return

    # 如果没有导出就导出
    with open("/etc/profile", "a") as f:
        last_line = lines[-1]
        if not last_line.endswith("\n"):
            # 说明最后一行没有换行，这个情况下先加上换行
            f.write("\n")
        f.write(export_str + "\n")


def export_header_files(pkg: str = None):
    """导出头文件

    Parameters:
    -----------
    pkg: str
        MySQL 安装包全路径

    Return:
    -------

    """
    logging.info(messages.FUN_STARTS.format(fname()))

    # 检查是否已经导出过了
    dst_include_dir = Path("/usr/include/") / "mysql-{}".format(
        get_mysql_version(pkg.name)
    )
    logging.info("dst_include_dir = mysql-{}".format(dst_include_dir))
    if dst_include_dir.exists():
        # 执行到这里说明已经导出过了
        logging.info("dst-incluce-dir '{}' exists. ".format(dst_include_dir))
        logging.info(messages.FUN_ENDS.format(fname()))
        return

    src_include_dir = Path(pkg_to_basedir(pkg)) / "include"
    logging.info("src_include_dir = {}".format(src_include_dir))

    # 复制 include 目录
    shutil.copytree(src_include_dir, dst_include_dir)

    # 结束
    logging.info(messages.FUN_ENDS.format(fname()))


def export_so_files(pkg: Path = None):
    """导出 so 文件

    Parameters:
    -----------
    pkg: str
        MySQL 安装包的大小

    """
    logging.info(messages.FUN_STARTS.format(fname()))

    conf_file = Path("/etc/ld.so.conf.d") / "mysql-{}.conf".format(
        get_mysql_version(pkg.name)
    )
    if not conf_file.exists():
        with open(conf_file, "w") as f:
            mysql_lib_dir = pkg_to_basedir(pkg) / "lib/"
            f.write(str(mysql_lib_dir))
            f.write("\n")

    exe_shell_cmd("ldconfig")

    logging.info(messages.FUN_ENDS.format(fname()))


def make_mysql_writable(port: int = 3306):
    """把结点设置为可写
    Parameters:
    -----------
    port: int
        MySQL 端口号

    Return
    ------
    None
    """
    logging.info(messages.FUN_STARTS.format(fname()))

    with dbma_mysql_cnx(port) as cursor:
        # sql = "set @@global.read_only=OFF; set @@global.super_read_only=OFF;"
        sql = "set @@global.read_only=OFF;"
        """
        有可能 MySQL 还没有启动、所以这里加一个重试 3 次的逻辑
        """
        retry_conts = 3
        for i in range(retry_conts):
            try:
                cursor.execute(sql)
                logging.info("make mysql instance 127.0.0.1:{} writable .".format(port))
                logging.info("sql = {}".format(cursor.statement))
                break
            except Exception as err:
                logging.exception(err)
                logging.warn("make mysql writeable failed '{}' ".format(str(err)))
                logging.warn(
                    "retry counts = {} , the max retry conts is {}", i, retry_conts - 1
                )

                # 加上 sleep 的逻辑
                j = i + 1
                sleep_time = 7 - (j * 2)
                logging.info("sleep({})".format(sleep_time))
                if sleep_time >= 1:
                    time.sleep(sleep_time)
                else:
                    time.sleep(1)

    logging.info(messages.FUN_ENDS.format(fname()))


def create_os_user_for_mysql(port: int = 3306):
    """
    创建 MySQL 在 OS 层面的用户

    Parameters:
    ----------
    port: int
        MySQL 的端口号


    """
    logging.info(messages.FUN_STARTS.format(fname()))
    logging.info("mysql-user 'mysql{}' ".format(port))
    user = MySQLUser(port)
    user.create()
    logging.info(messages.FUN_ENDS.format(fname()))

    return user


def create_directory(path: Path = None):
    """
    创建给定目录，如果你目录不存在就先创建父目录

    Parameters:
    -----------
    path: Path
        要创建的目录
    """
    logging.info(messages.FUN_STARTS.format(fname()))
    logging.info("path =  '{}' .".format(path))

    # 把 str 转成 Path
    if isinstance(path, str):
        path = Path(path)

    # 如果存在就直接退出
    if path.exists():
        return

    # 如果父目录都不存在就先创建父目录
    if not path.parent.exists():
        create_directory(path.parent)

    # 创建目录
    path.mkdir()

    logging.info(messages.FUN_ENDS.format(fname()))


def calculate_mysql_directory_by_port_and_type(
    port: int = 3306, dir_type: MySQLDirs = MySQLDirs.DATA
):
    """
    根据端口号和目录类型计算出对应的 Path 对象

    Parameters:
    -----------
    port:int
        MySQL 端口号

    dir_type: MySQLDirectorys
        MySQL 的目录类型
    """
    parent = None
    if dir_type == MySQLDirs.DATA:
        parent = Path(dbm_agent_config.mysql_datadir_parent)
    elif dir_type == MySQLDirs.BINLOG:
        parent = Path(dbm_agent_config.mysql_binlogdir_parent)
    elif dir_type == MySQLDirs.BACKUP:
        parent = Path(dbm_agent_config.mysql_backupdir_parent)
    else:
        raise NotSuportMySQLDirectoryType(
            "not suport MySQLDirectorys type '{}'".format(dir_type)
        )

    return parent / str(port)


def create_mysql_dirs(port: int = 3306, user: MySQLUser = None):
    """
    创建 MySQL 相关的目录(data/binlog/backup)

    Parameters:
    -----------
    port: int
        MySQL 端口号

    user: MySQLUser
        MySQL 在 os 层面的用户
    """
    for dir_type in MySQLDirs:
        path = calculate_mysql_directory_by_port_and_type(port, dir_type)
        create_directory(path)
        user.chown(path)



# -*- coding: utf-8 -*-

"""Resdis 自动化配置"""

import logging
from pathlib import Path
from jinja2 import Template
from datetime import datetime
from dataclasses import dataclass, asdict

from dbma.bil.fun import fname
from dbma.core import messages
from dbma.components.redis.exceptions import RedisConfigTemplateFileNotExistsException


@dataclass
class RedisConfig(object):
    """实现对 Redis 的配置文件生成

    Parameters
    ----------
    port : int
        Redis 要监控的端口
    """

    port: int = 6379
    dbfilename: str = "dump.rdb"
    loglevel: str = "notice"
    daemonize: str = "yes"
    # bind: str = "127.0.0.1 ::1"
    bind: str = "* -::*"
    redis_dir: str = "/database/redis/{}".format(port)
    protected_mode: str = "yes"
    tcp_backlog: int = 511
    unixsocketperm: str = "700"
    tcp_keepalive: int = 300
    now: str = datetime.now().isoformat()

    def __post_init__(self):
        self.redis_dir = "/database/redis/{}".format(self.port)

    def render_config(self) -> str:
        """渲染配置文件

        Returns
        -------
        str

        Excpetions:
        -----------
        RedisConfigTemplateFileNotExistsException 配置文件模板不存在的情况
        """
        logging.info(messages.FUN_STARTS.format(fname()))
        import dbma

        # 模板文件的位置
        redis_config__template_file = (
            Path(dbma.__file__).parent / "static/cnfs/redis.conf.jinja"
        )

        # 配置文件模板不存在就要报异常
        if not redis_config__template_file.exists():
            msg = "redis config template file '{}' not exists Excepiton ".format(
                str(redis_config__template_file)
            )
            logging.error(msg)
            raise RedisConfigTemplateFileNotExistsException(msg)

        # 执行到这里说明配置文件模板一定是存在的，准备渲染
        content = ""
        with open(redis_config__template_file) as file_object:
            content = file_object.read()

        # 渲染配置文件，并返回渲染之后的文本
        template = Template(content)
        logging.info(messages.FUN_ENDS.format(fname()))
        return template.render(asdict(self))

    def generate_config_file(self):
        """生成配置文件 /etc/redis-{self.port}.conf"""
        config_file = "/etc/redis-{}.conf".format(self.port)
        with open(config_file, "w") as config_object:
            content = self.render_config()
            config_object.write(content)


@dataclass
class RedisReplicaConfig(RedisConfig):
    """Redis 从结点配置文件生成类

    Parameters
    ----------
    RedisConfig :
    """

    is_replica: bool = True
    # format et: 127.0.0.1 6379
    replicaof: str = ""



# -*- coding: utf-8 -*-

"""代码写到一定的量之后发现，实在是太多的字符串常量了。另人难过的差不多的语境，不同时期的拼写还不一致。
作者: 蒋乐兴|neeky
时间: 2023-03
"""

# 函数开执行
FUN_STARTS = "starts {} ."
# 函数执行完成(return 之前)
FUN_ENDS = "ends {} ."
VIEW_FUN_STARTS = "view-request-starts: {}"
VIEW_FUN_ENDS = "view-requests-ends: {}"


# 文件不存在
FILE_NOT_EXISTS = "file '{}' not exists ."


# 执行命令
EXECUTE_CMD = "excute cmd '{}' ."


# 创建目录
CREATE_DIR = "create dir '{}' ."
DIR_EXISTS = "dir '{}' exists ."
DIR_NOT_EXISTS = "dir '{}' not exists ."
MOVE_FILE_TO = "move '{}' to '{}' "


# MySQL MESSAGES
MYSQL_INSTANCE_HAS_EXISTS = "mysql instance '{}' has exists ."


ARG_NOT_IN_POST_DICT = "argument '{}' not in post dict"



# -*- coding: utf-8 -*-

"""MySQL 安装相关的逻辑实现

作者: 蒋乐兴|neeky@live.com
时间: 2023-03
"""

import os
import shutil
import tarfile
import logging
from pathlib import Path
from datetime import datetime
from dbma.bil.fun import fname
from dbma.core import messages
from dbma.bil.osuser import MySQLUser
from dbma.bil.cmdexecutor import exe_shell_cmd
from dbma.core.configs import dbm_agent_config
from dbma.components.mysql.config import MySQLConfig
from dbma.components.mysql.commons import get_mysql_version
from dbma.components.mysql.commons import export_cmds_to_path
from dbma.components.mysql.commons import pkg_to_basedir, default_pkg
from dbma.components.mysql.commons import (
    export_header_files,
    export_so_files,
    create_os_user_for_mysql,
    create_mysql_dirs,
)
from dbma.components.mysql.exceptions import MySQLSystemdFileNotExists
from dbma.components.mysql.exceptions import MySQLPkgFileNotExistsException
from dbma.components.mysql.exceptions import InstanceHasBeenInstalledException
from dbma.components.mysql.asserts import (
    assert_mysql_install_pkg_exists,
    assert_mysql_datadir_not_exists,
    assert_mysql_systemd_file_exists,
    assert_mysql_systemd_file_not_exists,
)


def create_init_sql_file(version: str = None):
    """生成 init-sql 文件给 init 的时候用

    Parameters:
    -----------

    version: str
        MySQL 版本号

    Exception:
    ----------
    ValueError
    """
    logging.info(messages.FUN_STARTS.format(fname()))

    import dbma

    if version is None:
        message = "mysql version is None, can't create init-sql-file ."
        logging.error(message)
        raise ValueError(messages)

    # 检查版本号并根据版本号生成配置文件
    if version.startswith("8.0"):
        sql_file = Path(dbma.__file__).parent / "static/cnfs/init-8.0.x.sql"
    elif version.startswith("5.7"):
        sql_file = Path(dbma.__file__).parent / "static/cnfs/init-5.7.x.sql"
    else:
        message = (
            "mysql version is '{}' not suported, can't create init-sql-file .".format(
                version
            )
        )
        logging.error(message)
        raise ValueError(messages)
    logging.info("init sql file = {}".format(sql_file))

    # 复制文件
    shutil.copy(sql_file, dbm_agent_config.mysql_init_user_sql_file)
    logging.info(messages.FUN_ENDS.format(fname()))


def remove_init_sql_file():
    """清理 /tmp/mysql-init.sql 文件"""
    logging.info(messages.FUN_STARTS.format(fname()))

    init_sql_file = Path(dbm_agent_config.mysql_init_user_sql_file)
    if init_sql_file.exists():
        os.remove(init_sql_file)

    logging.info(messages.FUN_ENDS.format(fname()))


def checks_for_install(port: int = 3306, pkg: Path = default_pkg):
    """MySQL 安装前的检查

    Parameters:
    -----------
    port: int
        MySQL 端口

    pkg: Path
        MySQL 安装包的全路径

    Return:
    -------
    None

    Exceptions:
    ----------
    MySQLPkgFileNotExistsException
    MySQLDataDirectoryExists
    """
    logging.info(messages.FUN_STARTS.format(fname()))

    # 检查安装包是否存在
    assert_mysql_install_pkg_exists(pkg)

    # 检查给定的实例是不是已经安装过了
    assert_mysql_datadir_not_exists(port)

    logging.info(messages.FUN_ENDS.format(fname()))


def check_mysql_systemd_exists(port: int = 3306):
    """检查 mysql 的 systemd 配置文件是否存在

    Paramters:
    ----------
    port: int
        MySQL 端口号

    Returns:
    --------
    None

    Exceptions:
    -----------
    MySQLSystemdFileNotExists
    """
    logging.info(messages.FUN_STARTS.format(fname()))

    systemd_file = Path("/usr/lib/systemd/system/") / "mysqld-{}.service".format(port)
    if not systemd_file.exists():
        logging.error(messages.FILE_NOT_EXISTS.format(systemd_file))
        raise MySQLSystemdFileNotExists(systemd_file)

    logging.info(messages.FUN_ENDS.format(fname()))


def enable_systemd_for_mysql(port: int = 3306):
    """启用 systemd 服务

    Parameters:
    -----------
    port: int
        MySQL 端口号

    Return:
    -------
    None

    Exceptions:
    -----------
    MySQLSystemdFileNotExists
    """
    # 如果 systemd 配置文件不存在就报异常
    logging.info(messages.FUN_STARTS.format(fname()))

    try:
        # check_mysql_systemd_exists(port)
        assert_mysql_systemd_file_exists(port)
        # 没有报异常，说明 systemd 配置存在, 准备执行 enable 操作
        enable_cmd = "systemctl enable mysqld-{}".format(port)
        logging.info(messages.EXECUTE_CMD.format(enable_cmd))
        exe_shell_cmd(enable_cmd)
    except MySQLSystemdFileNotExists as err:
        raise err

    logging.info(messages.FUN_ENDS.format(fname()))


def disable_systemd_for_mysql(port: int = 3306):
    """禁用 systemd 服务

    Parameters:
    -----------
    port: int
        MySQL 端口号

    Return:
    -------
    None

    Exceptions:
    -----------
    MySQLSystemdFileNotExists
    """
    # 如果 systemd 配置文件不存在就报异常
    logging.info(messages.FUN_STARTS.format(fname()))

    try:
        # check_mysql_systemd_exists(port)
        assert_mysql_systemd_file_exists(port)
        # 没有报异常，说明 systemd 配置存在
        # 执行 enable 操作
        disable_cmd = "systemctl disable mysqld-{}".format(port)
        logging.info(messages.EXECUTE_CMD.format(disable_cmd))
        exe_shell_cmd(disable_cmd)
    except MySQLSystemdFileNotExists as err:
        logging.info(err)
        raise err

    logging.info(messages.FUN_ENDS.format(fname()))


def start_mysql(port: int = 3306):
    """启动 MySQL 数据库

    Parameter:
    ----------
    port: int
        MySQL 端口号

    Returns:
    -------
    None

    Exceptions:
    -----------
    MySQLSystemdFileNotExists
    """
    # 如果 systemd 配置文件不存在就报异常
    logging.info(messages.FUN_STARTS.format(fname()))

    try:
        # check_mysql_systemd_exists(port)
        assert_mysql_systemd_file_exists(port)
        # 没有报异常，说明 systemd 配置存在
        # 执行 start 操作
        start_cmd = "systemctl start mysqld-{}".format(port)
        logging.info(messages.EXECUTE_CMD.format(start_cmd))
        exe_shell_cmd(start_cmd)
    except MySQLSystemdFileNotExists as err:
        raise err

    logging.info(messages.FUN_ENDS.format(fname()))


def stop_mysql(port: int = 3306):
    """关闭 MySQL 数据库

    Parameter:
    ----------
    port: int
        MySQL 端口号

    Returns:
    -------
    None

    Exceptions:
    -----------
    MySQLSystemdFileNotExists
    """
    # 如果 systemd 配置文件不存在就报异常
    logging.info(messages.FUN_STARTS.format(fname()))

    try:
        # check_mysql_systemd_exists(port)
        assert_mysql_systemd_file_exists(port)
        # 没有报异常，说明 systemd 配置存在
        # 执行 stop 操作
        stop_cmd = "systemctl stop mysqld-{}".format(port)
        logging.info(messages.EXECUTE_CMD.format(stop_cmd))
        exe_shell_cmd(stop_cmd)
    except MySQLSystemdFileNotExists as err:
        raise err

    logging.info(messages.FUN_ENDS.format(fname()))


def create_user_and_dirs(port: int = 3306):
    """根据端口号创建 MySQL 用户

    Parameter:
    ----------
    port: int
        MySQL 的端口号

    Return:
    -------
        None
    """
    logging.info(messages.FUN_STARTS.format(fname()))

    user = create_os_user_for_mysql(port)
    create_mysql_dirs(port, user)

    # # 创建用户
    # user = MySQLUser(port)
    # user.create()

    # # 创建 datadir & binlogdir
    # datadir = Path(dbm_agent_config.mysql_datadir_parent) / str(port)
    # binlogdir = Path(dbm_agent_config.mysql_binlogdir_parent) / str(port)

    # if not datadir.exists():
    #     logging.info(messages.CREATE_DIR.format(datadir))
    #     os.mkdir(datadir)
    # else:
    #     logging.warning(messages.DIR_EXISTS.format(datadir))

    # if not binlogdir.exists():
    #     logging.info(messages.CREATE_DIR.format(binlogdir))
    #     os.mkdir(binlogdir)
    # else:
    #     logging.warning(messages.DIR_EXISTS.format(binlogdir))

    # user.chown(datadir)
    # user.chown(binlogdir)

    logging.info(messages.FUN_ENDS.format(fname()))


def backup_dirs(port: int = 3306, suffix=None):
    """备份数据目录和 binlog 目录，如果 suffix 为 None 就用当前时间值

    Parameters:
    -----------
    port: int
        MySQL 端口号

    suffix: str
        备份文件后缀

    Returns:
    --------
    None
    """
    logging.info(messages.FUN_STARTS.format(fname()))

    # 计算 suffix
    if suffix is None:
        suffix = datetime.now().isoformat().replace(":", "-").replace(".", "-")

    # 计算当前的 datadir 和 binlogdir
    datadir = Path(dbm_agent_config.mysql_datadir_parent) / str(port)
    binlogdir = Path(dbm_agent_config.mysql_binlogdir_parent) / str(port)

    datadir_backup_dir = "{}-backup-{}".format(datadir, suffix)
    binlogdir_backup_dir = "{}-backup-{}".format(binlogdir, suffix)

    shutil.move(datadir, datadir_backup_dir)
    shutil.move(binlogdir, binlogdir_backup_dir)

    logging.info(messages.FUN_ENDS.format(fname()))


def backup_config_file(port: int = 3306, suffix=None):
    """备份配置文件

    Parameters:
    -----------
    port: int
        MySQL 端口号

    suffix: str
        备份文件的后缀

    Return:
    -------

    """
    logging.info(messages.FUN_STARTS.format(fname()))

    # 计算 suffix
    if suffix is None:
        suffix = datetime.now().isoformat().replace(":", "-").replace(".", "-")
    datadir = Path(dbm_agent_config.mysql_datadir_parent) / str(port)
    if not datadir.exists():
        logging.warning(messages.DIR_NOT_EXISTS.format(datadir))
        return

    config_file = "/etc/my-{}.cnf".format(port)
    config_backup_file = datadir / "my-{}.cnf-backup-{}".format(port, suffix)
    logging.info(messages.MOVE_FILE_TO.format(config_file, config_backup_file))

    # 备份文件
    shutil.copyfile(config_file, config_backup_file)
    os.remove(config_file)

    logging.info(messages.FUN_ENDS.format(fname()))


def decompression_pkg(pkg: Path = default_pkg):
    """解压安装包到 /usr/local/

    Parameter:
    ---------
    pkg: Path
        MySQL 安装包的全路径

    Return:
    -------
        None

    Exception:
    ---------
        MySQLPkgFileNotExistsException
    """
    logging.info(messages.FUN_STARTS.format(fname()))

    if not pkg.exists():
        logging.error(messages.FILE_NOT_EXISTS.format(pkg))
        raise MySQLPkgFileNotExistsException(messages.FILE_NOT_EXISTS.format(pkg))

    # TODO
    # / 运行加大了测试用例的编写难度
    # 这个后面改掉
    basedir = pkg_to_basedir(pkg)
    flag_file = basedir / ".dbm-agent-decompression.txt"
    if flag_file.exists():
        logging.info("ends decompression pkg .")
        return

    # 准备解压
    with tarfile.open(pkg) as tar_pkg:
        tar_pkg.extractall("/usr/local/")

    # 解压完成之后写入标记文件 basedir/.dbm-agent-decompression.txt
    with open(flag_file, "w") as f:
        f.write("dbm-agent")

    # 解压完成
    logging.info(messages.FUN_ENDS.format(fname()))


def create_mysql_config_file(
    port: int = 3306, basedir: Path = None, innodb_buffer_pool_size: str = "128M"
):
    """创建 MySQL 配置文件

    Parameter:
    ----------
    port: int
        MySQL 端口号、默认值 3306

    basedir: Path
        MySQL basedir

    innodb_buffer_pool_size: int
        缓冲池的大小

    Return:
    -------
        None
    """
    logging.info(messages.FUN_STARTS.format(fname()))
    logging.info(
        "basedir = '{}', port = '{}', innodb_buffer_pool_size = '{}' .".format(
            basedir, port, innodb_buffer_pool_size
        )
    )

    config = MySQLConfig(
        basedir=str(basedir), port=port, innodb_buffer_pool_size=innodb_buffer_pool_size
    )

    config.calcu_second_attrs()
    config.generate_cnf_config_file()
    config.generate_init_cnf_config_file()
    config.generate_systemd_cnf_config()

    logging.info(messages.FUN_ENDS.format(fname()))


def init_mysql(port: int = 3306, basedir: Path = None):
    """初始化 MySQL

    Parameter:
    ----------
    port: int
        MySQL 端口号

    basedir: Path
        MySQL basedir

    Return:
    -------
        None
    """
    logging.info(messages.FUN_STARTS.format(fname()))
    logging.info("port = '{}', basedir = '{}' .".format(port, basedir))

    mysqld = Path(basedir) / "bin/mysqld"
    config = Path(dbm_agent_config.mysql_init_cnf)

    init_cmd = "{} --defaults-file={} --init-file={} --initialize-insecure".format(
        mysqld, config, dbm_agent_config.mysql_init_user_sql_file
    )
    logging.info("init-cmd = '{}' .".format(init_cmd))
    # 执行 init 操作
    exe_shell_cmd(init_cmd)

    logging.info(messages.FUN_ENDS.format(fname()))


def install_mysql(
    port: int = 3306, pkg: Path = None, innodb_buffer_pool_size: str = "128M"
):
    """安装 MySQL 实例

    Parameter:
    ----------
    port: int
        MySQL 端口号

    pkg: Path
        MySQL 安装包全路径

    Returns:
    --------
    None

    Exceptions:
    -----------
    MySQLPkgFileNotExistsException

    InstanceHasBeenInstalledException
    """
    logging.info(messages.FUN_STARTS.format(fname()))

    # 安装前的检查
    try:
        checks_for_install(port, pkg)
    except MySQLPkgFileNotExistsException as err:
        logging.error("pkg not exists, skip install . {} .".format(err))
        raise err
    except InstanceHasBeenInstalledException as err:
        logging.error("instance has been installed {} .".format(err))
        raise err
    except Exception as err:
        logging.info("unknown Exception {}".format(err))
        raise err

    version = get_mysql_version(pkg.name)

    # 第一步 创建用户和目录
    create_user_and_dirs(port)

    # 第二步 解压安装包
    decompression_pkg(pkg)

    # 第三步 计算 basedir
    basedir = pkg_to_basedir(pkg)

    # 第四步 创建配置文件
    create_mysql_config_file(
        port=port, basedir=basedir, innodb_buffer_pool_size=innodb_buffer_pool_size
    )

    # 第五步 复制 init 文件
    create_init_sql_file(version)

    # 第五步 初始化 mysql 实例
    init_mysql(port=port, basedir=basedir)

    # 第六步 启用 mysqld 服务
    enable_systemd_for_mysql(port)

    # 第七步 启动 mysql 实例
    start_mysql(port)

    # 第八步 导出 PATH 环境变量
    export_cmds_to_path(basedir)

    # 第九步 导出头文件
    export_header_files(pkg)

    # 第十步 导出 so 文件
    export_so_files(pkg)

    # 清理 init-sql
    remove_init_sql_file()

    logging.info(messages.FUN_ENDS.format(fname()))


def uninstall_mysql(port: int = 3306):
    """卸载 mysql 数据库

    Exceptions:
    -----------

    MySQLSystemdFileNotExists
    """
    logging.info(messages.FUN_STARTS.format(fname()))

    # 第一步 停止 MySQL 服务
    stop_mysql(port)

    # 第二步 禁用 systemd 服务
    disable_systemd_for_mysql(port)

    # 第三步 备份配置文件
    backup_config_file(port)

    # 第四步 备份数据目录与binlog目录
    backup_dirs(port)
    logging.info(messages.FUN_ENDS.format(fname()))



# -*- coding: utf-8 -*-

"""MySQL 相关的一些断言项
"""

from pathlib import Path
from dbma.core.configs import dbm_agent_config
from dbma.components.mysql.exceptions import (
    MySQLPkgFileNotExistsException,
    MySQLDataDirectoryExists,
    MySQLSystemdFileExists,
    MySQLSystemdFileNotExists,
)


def assert_mysql_install_pkg_exists(pkg: Path = None):
    """
    断言 MySQL 的安装包文件是否存在，如果不存在就报异常

    parameters:
    -----------
    pkg: Path
        MySQL 安装包全路径

    Exceptions:
    -----------
    MySQLPkgFileNotExistsException
    """
    # 如果只是给了一个安装文件的名字，就把默认的前缀补上
    if pkg.is_absolute() and str(pkg).startswith("mysql"):
        pkg = Path("/usr/local/dbm-agent/pkgs/") / pkg

    # 不存在就报异常
    if not pkg.exists():
        raise MySQLPkgFileNotExistsException(
            "mysql install pkg '{}' not exists .".format(str(pkg))
        )


def assert_mysql_datadir_not_exists(port: int = 3306):
    """
    断言对应端口的 MySQL 实例的数据目录存在

    parameters:
    -----------
    port: int
        MySQL 端口号

    Exceptions:
    -----------
    MySQLDataDirectoryExists
    """
    datadir = Path(dbm_agent_config.mysql_datadir_parent) / "{}".format(port)
    if datadir.exists():
        message = "mysql data '{}' directory exists .".format(datadir)
        raise MySQLDataDirectoryExists(message)


def assert_mysql_systemd_file_not_exists(port: int = 3306):
    """
    断言对应端口的 MySQL systemd 配置文件不存在、如果实际的情况与断言不一致就报异常

    parameters:
    -----------
    port: int
        MySQL 端口号

    Exceptions:
    -----------
    MySQLSystemdFileExists
    """
    systemd_file = Path("/usr/lib/systemd/system/mysqld-{}.service".format(port))
    if systemd_file.exists():
        message = "mysql systemd config file '{}' exists.".format(str(systemd_file))
        raise MySQLSystemdFileExists(message)


def assert_mysql_systemd_file_exists(port: int = 3306):
    """
    断言对应端口的 MySQL systemd 配置文件存在、如果实际的情况与断言不一致就报异常

    parameters:
    -----------
    port: int
        MySQL 端口号

    Exceptions:
    -----------
    MySQLSystemdFileExists
    """
    systemd_file = Path("/usr/lib/systemd/system/mysqld-{}.service".format(port))
    if not systemd_file.exists():
        # 如果不存在就要报异常
        message = "mysql systemd config file '{}' exists.".format(str(systemd_file))
        raise MySQLSystemdFileNotExists(message)



import os
from setuptools import setup


def get_version():
    """
    动态获取 dbm-agent 的版本号
    """

    project_dir_name = os.path.dirname(__file__)
    version_file_path = os.path.join(project_dir_name, "dbma/version.py")
    with open(version_file_path) as version_file_obj:
        content = version_file_obj.read()

    g = {}

    exec(content, g, g)
    return g["VERSION"]


agent_version = get_version()

setup(
    name="dbm-agent",
    version=agent_version,
    description="dbm-agent 数据库管理中心客户端程序",
    author="Neeky",
    author_email="neeky@live.com",
    maintainer="Neeky",
    maintainer_email="neeky@live.com",
    scripts=[
        "bin/dbm-agent",
        "bin/dbma-cli-init",
        "bin/dbma-cli-mysql",
        "bin/dbma-cli-redis",
        "bin/dbm-bt-conn-stack",
    ],
    packages=[
        "dbma",
        "dbma/core",
        "dbma/core/views",
        "dbma/core/agent",
        "dbma/core/threads",
        "dbma/bil",
        "dbma/components",
        "dbma/components/mysql",
        "dbma/components/mysql/views",
        "dbma/components/mysql/backups",
        "dbma/components/mysql/backends",
        "dbma/components/redis",
        "dbma/components/orchestrator",
    ],
    package_data={"dbma": ["static/cnfs/*", "static/sql-scripts/*"]},
    url="https://github.com/Neeky/dbm-agent",
    install_requires=[
        "mysql-connector-python>=8.0.31",
        "redis>=4.5.4",
        "aiohttp>=3.8.1",
        "requests>=2.22.0",
        "Jinja2>=2.10.1",
        "psutil>=5.6.6",
        "distro>=1.4.0",
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Operating System :: POSIX",
        "Operating System :: MacOS :: MacOS X",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
)
